package java.util.concurrent;

import java.util.concurrent.locks.AbstractQueuedSynchronizer;

public class CountDownLatch
{
    private final CountDownLatch.Sync sync;
    
    public CountDownLatch(int n) {
        if (n < 0) {
            throw new IllegalArgumentException("count < 0");
        }
        this.sync = new Sync(n);
        return;
    }
    
    public void await() throws InterruptedException {
        this.sync.acquireSharedInterruptibly(1);
    }
    
    public boolean await(long n, TimeUnit timeUnit) throws InterruptedException {
        return this.sync.tryAcquireSharedNanos(1, timeUnit.toNanos(n));
    }
    
    public void countDown() {
        this.sync.releaseShared(1);
    }
    
    public long getCount() {
        return (long)this.sync.getCount();
    }
    
    @Override
    public String toString() {
        return super.toString() + "[Count = " + this.sync.getCount() + "]";
    }
    
    private static final class Sync extends AbstractQueuedSynchronizer
    {
        private static final long serialVersionUID = 4982264981922014374L;
        
        Sync(int state) {
            this.setState(state);
        }
        
        int getCount() {
            return this.getState();
        }
        
        @Override
        protected int tryAcquireShared(int n) {
            return (this.getState() == 0) ? 1 : -1;
        }
        
        @Override
        protected boolean tryReleaseShared(int n) {
            while (true) {
                int state = this.getState();
                if (state == 0) {
                    return false;
                }
                int update = state - 1;
                if (this.compareAndSetState(state, update)) {
                    return update == 0;
                }
                continue;
            }
        }
    }
}
